// This file was automatically generated by collect_driver_connection_prefix.py
// DO NOT EDIT

// clang-format off
constexpr static struct {
  const char* pszPrefix;
  const char* pszDriverName;
}
asKnownConnectionPrefixes[] =
{
   {"ADBC:", "ADBC"},
   {"AIVector:", "AIVector"},
   {"AMIGOCLOUD:", "AMIGOCLOUD"},
   {"CARTO:", "CARTO"},
   {"CSW:", "CSW"},
   {"DAAS:", "DAAS"},
   {"EEDA:", "EEDA"},
   {"EEDAI:", "EEDAI"},
   {"ES:", "Elasticsearch"},
   {"GMLAS:", "GMLAS"},
   {"GPSBABEL:", "GPSBABEL"},
   {"GTI:", "GTI"},
   {"HANA:", "HANA"},
   {"IDB:", "IDB"},
   {"MongoDBv3:", "MongoDBv3"},
   {"MSSQL:", "MSSQLSpatial"},
   {"MYSQL:", "MYSQL"},
   {"NGW:", "NGW"},
   {"NULL:", "NULL"},
   {"OAPIF:", "OAPIF"},
   {"OCI:", "OCI"},
   {"ODBC:", "ODBC"},
   {"PG:", "PostgreSQL"},
   {"PLMOSAIC:", "PLMOSAIC"},
   {"PLSCENES:", "PLSCENES"},
   {"WFS:", "WFS"},
   {"WMTS:", "WMTS"},
};

// clang-format on
