/* This file has been generated by generate_character_set_conv_tables.c */
/* DO NOT EDIT !*/

/* clang-format off */
typedef unsigned char CPLCodePageConvTable[128][3];

const CPLCodePageConvTable* CPLGetConversionTableToUTF8(const char* pszEncoding);
/* clang-format on */
