/******************************************************************************
 *
 * Project:  OpenGIS Simple Features Reference Implementation
 * Purpose:  Defines OGRLayerWithTranslateFeature class
 * Author:   Even Rouault, even dot rouault at spatialys.com
 *
 ******************************************************************************
 * Copyright (c) 2025, Even Rouault <even dot rouault at spatialys.com>
 *
 * SPDX-License-Identifier: MIT
 ****************************************************************************/

#include "ogrlayerwithtranslatefeature.h"

#ifndef DOXYGEN_SKIP

OGRLayerWithTranslateFeature::~OGRLayerWithTranslateFeature() = default;

#endif
